
//
// AXI Dual-Ported memory
// 
// This is intended for testbenches but can be synthesized.
// FPGA compilers will map the behvioural memory to real DPR.
// (At least Xilinx will....)
// 
// This AXI memory has been developed alongside the axi_mux code
// to test the multiplexing of multitple AXI streams.
// As with most test benches I had to balance the time spend on
// the test bench versus the actual code being tested. 
// Therefore this memory is not fully parameterised
// This it for now it is only 32 bits wide... 
//
// To have great flexibility on the inputs
// I put some small FIFOs on every part of the AXI bus
// This has the disadvantadge that any read or write cycle
// will incure a once cycle latency, as will the responses.
//

module axi_dpm32
#( parameter ADRS_BITS = 32,
             DATA_BITS = 32,
             LEN_BITS  = 4,
             ID_BITS   = 6,
             // 2-Log depth of the memory in data bits above 
           L2MEM_DEPTH = 10
 )             
(
   input              clk    ,
   input              reset_n,
   
   //
   // AXI port 
   //
   
   // write address 
   
   input                    awvalid , 
   output                   awready , 
   input    [ADRS_BITS-1:0] awaddr  , 
   input     [LEN_BITS-1:0] awlen   , 
   input             [ 2:0] awsize  , 
   input      [ID_BITS-1:0] awid    , 
   input             [ 1:0] awburst , 
                           
   // write data           
   input                    wvalid  , 
   output                   wready  , 
   input    [DATA_BITS-1:0] wdata   , 
   input  [DATA_BITS/8-1:0] wstrb   , 
   input                    wlast   , 
   input      [ID_BITS-1:0] wid     , 

   // write response    
   output                   bvalid  , 
   input                    bready  , 
   output           [ 1 :0] bresp   , 
   output     [ID_BITS-1:0] bid     , 


   //
   // read interface
   //
   
   // read address 
   input                    arvalid, 
   output                   arready, 
   input    [ADRS_BITS-1:0] araddr , 
   input      [ID_BITS-1:0] arid   , 
   input            [ 3: 0] arlen  , 
   input            [ 2: 0] arsize , 
   input            [ 1: 0] arburst, 

   // read data
   output                   rvalid , 
   input                    rready , 
   output  [DATA_BITS-1:0]  rdata  , 
   output                   rlast  , 
   output           [ 1 :0] rresp  , 
   output     [ID_BITS-1:0] rid    
);

localparam MEMDEPTH = (1<<L2MEM_DEPTH);



 /*****************************\                        
  *   _    _      _ _         *
  *  | |  | |    (_) |        *
  *  | |  | |_ __ _| |_ ___   *
  *  | |/\| | '__| | __/ _ \  *
  *  \  /\  / |  | | ||  __/  *
  *   \/  \/|_|  |_|\__\___|  *
  *                           *
 \*****************************/
 
// Write address (AW)
// Using address, length, size, ID and burst 
localparam AW_WIDTH = L2MEM_DEPTH+LEN_BITS+3+ID_BITS+2;

wire                awff_write;
wire [AW_WIDTH-1:0] awff_wdata;
reg                 awff_read;
wire [AW_WIDTH-1:0] awff_rdata;
wire                awff_full;
wire                awff_empty;


   // Use WORD part of address 
   assign awff_wdata = {awaddr[L2MEM_DEPTH+1:2],awlen,
                        awsize,awid,awburst};
                        
   assign awff_write = awvalid & awready;

 sync_fifo
#(.WIDTH(AW_WIDTH),     // width in bits
   .L2DEPTH(4),         // Log 2 Depth, 4=16 deep
   .REGFLAGS(0)         // Full, empty are registered
)
aw_fifo
(
   .clk    (clk      ),   // system clock                 
   .reset_n(reset_n  ), // A-synchronous low reset/clear
   .enable (1'b1     ), // clock gating                 
   .clear  (1'b0     ), // Synchronous clear            
                                            
   .write  (awff_write), // write FIFO                   
   .wdata  (awff_wdata), // write data                   
   .read   (awff_read ), // read FIFO                    
   .rdata  (awff_rdata), // read data                    
         
   .empty  (awff_empty), // FIFO is empty                
   .full   (awff_full ), // FIFO is full                 
   .level  ()            // Fill level                   
);

   assign awready  = !awff_full;

   // Post FIFO signals 
   
wire [L2MEM_DEPTH-1:0] i_awadrs ; 
wire    [LEN_BITS-1:0] i_awlen  ; 
wire            [ 2:0] i_awsize ; 
wire     [ID_BITS-1:0] i_awid   ; 
wire            [ 1:0] i_awburst; 

   assign { i_awadrs ,
            i_awlen  ,
            i_awsize ,
            i_awid   ,
            i_awburst} = awff_rdata;

//            
// Write data (W)
// Using data, strobes, ID and last 
//

localparam W_WIDTH = DATA_BITS+DATA_BITS/8+1+ID_BITS;

wire               wff_write;
wire [W_WIDTH-1:0] wff_wdata;
reg                wff_read;
wire [W_WIDTH-1:0] wff_rdata;
wire               wff_full;
wire               wff_empty;

   assign wff_wdata = {wdata,wstrb,wlast,wid};
                        
   assign wff_write = wvalid & wready;

 sync_fifo
#(.WIDTH(W_WIDTH),     // width in bits
   .L2DEPTH(6),         // Log 2 Depth, 6=64 deep
   .REGFLAGS(0)         // Full, empty are registered
)
w_fifo
(
   .clk    (clk      ),   // system clock                 
   .reset_n(reset_n  ), // A-synchronous low reset/clear
   .enable (1'b1     ), // clock gating                 
   .clear  (1'b0     ), // Synchronous clear            
                                            
   .write  (wff_write), // write FIFO                   
   .wdata  (wff_wdata), // write data                   
   .read   (wff_read ), // read FIFO                    
   .rdata  (wff_rdata), // read data                    
         
   .empty  (wff_empty), // FIFO is empty                
   .full   (wff_full ), // FIFO is full                 
   .level  ()           // Fill level                   
);

   assign wready  = !wff_full;

wire    [DATA_BITS-1:0] i_wdata; 
wire  [DATA_BITS/8-1:0] i_wstrb; 
wire                    i_wlast; 
wire      [ID_BITS-1:0] i_wid  ; 

   assign {i_wdata,
           i_wstrb,
           i_wlast,
           i_wid  } = wff_rdata;

//
// Write response (B)
//
localparam B_WIDTH = ID_BITS+2;

wire               bff_write;
wire [B_WIDTH-1:0] bff_wdata;
wire               bff_read;
wire [B_WIDTH-1:0] bff_rdata;
wire               bff_full;
wire               bff_empty;


   // Give respsonse when we read last data burst from FIFO  
   assign bff_write = wff_read & i_wlast;
   
   always @ (posedge clk)
      if (bff_write & bff_full)
      begin
         $display("%m @%0t: write to full response FF.\nStopping...",$time);
         #1 $stop;
      end
   
   // Read uses valid & ready 
   assign bff_read = bvalid & bready;
   
   // always give OK response mainly because I am lazy
   // You see: error handling is very tricky in this code
   // as the errors are detected with the address but a
   // error bresponse requires flushing part of the write data fifo.
   assign bff_wdata = {2'b00,i_wid};

// Depth is same as address FIFO 
 sync_fifo
#(.WIDTH(B_WIDTH),     // width in bits
   .L2DEPTH(4),         // Log 2 Depth, 4=16 deep
   .REGFLAGS(0)         // Full, empty are registered
)
b_fifo
(
   .clk    (clk      ),   // system clock                 
   .reset_n(reset_n  ), // A-synchronous low reset/clear
   .enable (1'b1     ), // clock gating                 
   .clear  (1'b0     ), // Synchronous clear            
                                            
   .write  (bff_write), // write FIFO                   
   .wdata  (bff_wdata), // write data                   
   .read   (bff_read ), // read FIFO                    
   .rdata  (bff_rdata), // read data                    
         
   .empty  (bff_empty), // FIFO is empty                
   .full   (bff_full ), // FIFO is full                 
   .level  ()            // Fill level                   
);

   assign bvalid  = !bff_empty;
   assign {bresp,bid} = bff_rdata;
           
   //
   // Do the writes to the memory 
   //
   
localparam W_IDLE = 1'b0,
            W_BURST = 1'b1;
            
reg   [1:0] write_state,nxt_write_state; 
reg         mem_write;

reg  [L2MEM_DEPTH-1:0] write_adrs,nxt_write_adrs;
reg     [LEN_BITS-1:0] write_count,nxt_write_count; 

   
   always @( * )
   begin
      // Defaults:
      mem_write = 1'b0;
      awff_read = 1'b0;
      wff_read  = 1'b0;
      nxt_write_state = write_state;
      nxt_write_count = write_count;
      nxt_write_adrs  = write_adrs;
      
      
      case (write_state)
      W_IDLE:
         if (!awff_empty & !wff_empty)
         begin
            mem_write = 1'b1;
            write_adrs = i_awadrs;
            wff_read  = 1'b1;
            
            // single beat or multi beat?
            if (i_awlen==0)
            begin
               awff_read = 1'b1;
               nxt_write_state = W_IDLE;
               
`ifdef AXI_PROT_CHECK_ON               
               if (!i_wlast)
               begin
                  $display("%m,@%0t: AXIWERR missing wlast.\nStopping...",$time);
                   #1 $stop;                  
               end
               
`endif // AXI_PROT_CHECK_ON               
            end
            else
            begin
               nxt_write_count = i_awlen-1;
               if (i_awburst==2'b01)
                  nxt_write_adrs = i_awadrs + 1; // word adrs increment
              else
                  nxt_write_adrs = i_awadrs;
               
               nxt_write_state = W_BURST;
               
`ifdef AXI_PROT_CHECK_ON               
               if (i_wlast)
               begin
                  $display("%m,@%0t: AXIWERR early wlast.\nStopping...",$time);
                   #1 $stop;                  
               end
`endif // AXI_PROT_CHECK_ON  
            end // multi beat
 
`ifdef AXI_PROT_CHECK_ON               
             if ( (i_awsize==1) && i_awadrs[  0]!=1'h0 ) |
                  (i_awsize==2) && i_awadrs[1:0]!=2'h0 ) 
                )
             begin
               $display("%m,@%0t: AXIWERR size/address mismatch.\nStopping...",$time);
                #1 $stop;                  
            end
`endif // AXI_PROT_CHECK_ON  

          end
          
          
      W_BURST: 
         if (!wff_empty)
         begin
            mem_write = 1'b1;
            wff_read  = 1'b1;
            nxt_write_count = write_count-1;
            if (i_awburst==2'b01) // incrementing
               nxt_write_adrs = write_adrs + 1; // word adrs increment
            else
               nxt_write_adrs = write_adrs; 

            
            if (write_count==0)
            begin
               awff_read = 1'b1;
               nxt_write_state = W_IDLE;

`ifdef AXI_PROT_CHECK_ON               
               if (!i_wlast)
               begin
                  $display("%m,@%0t: AXIWERR missing wlast.\nStopping...",$time);
                   #1 $stop;                  
               end
               
`endif // AXI_PROT_CHECK_ON               
               end
            else
            begin
            
`ifdef AXI_PROT_CHECK_ON               
               if (i_wlast)
               begin
                  $display("%m,@%0t: AXIWERR early wlast.\nStopping...",$time);
                   #1 $stop;                  
               end
`endif // AXI_PROT_CHECK_ON  
            end
            
         end
         
      endcase

   end
   
   always @(posedge clk or negedge reset_n)
   begin
      if (!reset_n)
      begin
         write_state <= W_IDLE;
         write_adrs  <= 0;
         write_count <= 0;
      end
      else
      begin
         write_state <= nxt_write_state;
         write_adrs  <= nxt_write_adrs;
         write_count <= nxt_write_count;
      end
   end

   
  /*****************************\
   *  ______               _   *
   *  | ___ \             | |  *
   *  | |_/ /___  __ _  __| |  *
   *  |    // _ \/ _` |/ _` |  *
   *  | |\ \  __/ (_| | (_| |  *
   *  \_| \_\___|\__,_|\__,_|  *
   *                           *
  \*****************************/

// Read address (AR)
// Using address, length, size, ID and burst 
localparam AR_WIDTH = L2MEM_DEPTH+LEN_BITS+3+ID_BITS+2;

wire                arff_write;
wire [AR_WIDTH-1:0] arff_wdata;
reg                 arff_read;
wire [AR_WIDTH-1:0] arff_rdata;
wire                arff_full;
wire                arff_empty;


   // Use WORD part of address 
   assign arff_wdata = {araddr[L2MEM_DEPTH+1:2],arlen,
                        arsize,arid,arburst};
                        
   assign arff_write = arvalid & arready;

 sync_fifo
#(.WIDTH(AR_WIDTH),     // width in bits
   .L2DEPTH(4),         // Log 2 Depth, 4=16 deep
   .REGFLAGS(0)         // Full, empty are registered
)
ar_fifo
(
   .clk    (clk      ),   // system clock                 
   .reset_n(reset_n  ), // A-synchronous low reset/clear
   .enable (1'b1     ), // clock gating                 
   .clear  (1'b0     ), // Synchronous clear            
                                            
   .write  (arff_write), // write FIFO                   
   .wdata  (arff_wdata), // write data                   
   .read   (arff_read ), // read FIFO                    
   .rdata  (arff_rdata), // read data                    
         
   .empty  (arff_empty), // FIFO is empty                
   .full   (arff_full ), // FIFO is full                 
   .level  ()            // Fill level                   
);

   assign arready  = !arff_full;

   // Post FIFO signals 
   
wire [L2MEM_DEPTH-1:0] i_aradrs ; 
wire    [LEN_BITS-1:0] i_arlen  ; 
wire            [ 2:0] i_arsize ; 
wire     [ID_BITS-1:0] i_arid   ; 
wire            [ 1:0] i_arburst; 

   assign { i_aradrs ,
            i_arlen  ,
            i_arsize ,
            i_arid   ,
            i_arburst} = arff_rdata;

//
// Read response (R)
// Storing data, id, last & response 
localparam R_WIDTH = DATA_BITS+ID_BITS+1+2,
           R_L2DEPTH = 4,
           R_DEPTH   = (1<<R_L2DEPTH);

reg                rff_write;
wire [R_WIDTH-1:0] rff_wdata;
wire               rff_read;
wire [R_WIDTH-1:0] rff_rdata;
wire               rff_full;
wire               rff_empty;
wire [R_L2DEPTH:0] rff_level;

// Read response data
reg [DATA_BITS-1:0] i_rdata;
reg   [ID_BITS-1:0] i_rid;
reg           [1:0] i_rresp;
reg                 i_rlast;
   
   always @ (posedge clk)
      if (rff_write & rff_full)
      begin
         $display("%m @%0t: write to full read data FF.\nStopping...",$time);
         #1 $stop;
      end
   
   // Read uses valid & ready 
   assign rff_read = rvalid & rready;
   
   assign rff_wdata = {i_rresp,i_rid,i_rlast,i_rdata};

// We can control our read speed
// Thus don't need a deep FIFO 
 sync_fifo
#(.WIDTH(R_WIDTH),      // width in bits
   .L2DEPTH(R_L2DEPTH), // Log 2 Depth, 3=8 deep
   .REGFLAGS(0)         // Full, empty are registered
)
r_fifo
(
   .clk    (clk      ),   // system clock                 
   .reset_n(reset_n  ), // A-synchronous low reset/clear
   .enable (1'b1     ), // clock gating                 
   .clear  (1'b0     ), // Synchronous clear            
                                            
   .write  (rff_write), // write FIFO                   
   .wdata  (rff_wdata), // write data                   
   .read   (rff_read ), // read FIFO                    
   .rdata  (rff_rdata), // read data                    
         
   .empty  (rff_empty), // FIFO is empty                
   .full   (rff_full ), // FIFO is full                 
   .level  (rff_level)  // Fill level                   
);

   assign rvalid  = !rff_empty;
   assign {rresp,rid,rlast,rdata} = rff_rdata;

   //
   // Do the reads from the memory 
   //
   
localparam R_IDLE = 1'b0,
           R_BURST = 1'b1;
            
reg   [1:0] read_state,nxt_read_state; 
reg         mem_read;

reg  [L2MEM_DEPTH-1:0] read_adrs,nxt_read_adrs;
reg     [LEN_BITS-1:0] read_count,nxt_read_count; 

   
   always @( * )
   begin
      // Defaults:
      mem_read  = 1'b0;
      arff_read = 1'b0;
      nxt_read_state = read_state;
      nxt_read_count = read_count;
      nxt_read_adrs  = read_adrs;
      
      
      case (read_state)
      R_IDLE:
         if (!arff_empty && rff_level<(R_DEPTH-2))
         begin
            mem_read  = 1'b1;
            read_adrs = i_aradrs;
            
            // single beat or multi beat?
            if (i_arlen==0)
            begin
               arff_read = 1'b1;
               nxt_read_state = R_IDLE;
            end
            else
            begin
               nxt_read_count = i_arlen-1;
               if (i_arburst==2'b01)
                  nxt_read_adrs = i_aradrs + 1; // word adrs increment
              else
                  nxt_read_adrs = i_aradrs;
               
               nxt_read_state = R_BURST;
               
            end // multi beat
 
`ifdef AXI_PROT_CHECK_ON               
             if ( (i_arsize==1) && i_aradrs[  0]!=1'h0 ) |
                  (i_arsize==2) && i_aradrs[1:0]!=2'h0 ) 
                )
             begin
               $display("%m,@%0t: AXIRERR size/address mismatch.\nStopping...",$time);
                #1 $stop;                  
            end
`endif // AXI_PROT_CHECK_ON  

          end
          
          
      R_BURST: 
         if (rff_level<(R_DEPTH-2))
         begin
            mem_read = 1'b1;
            nxt_read_count = read_count-1;
            if (i_arburst==2'b01) // incrementing
               nxt_read_adrs = read_adrs + 1; // word adrs increment
            else
               nxt_read_adrs = read_adrs; 

            if (read_count==0)
            begin
               arff_read = 1'b1;
               nxt_read_state = R_IDLE;
            end
            
         end
         
      endcase

   end
   
   always @(posedge clk or negedge reset_n)
   begin
      if (!reset_n)
      begin
         rff_write  <= 1'b0;
         read_state <= R_IDLE;
         read_adrs  <= 0;
         read_count <= 0;         
      end
      else
      begin
         // Write return data FIFO after a memory read 
         // Also set other return data as it is
         // picked up one cycle later          
         rff_write <= mem_read;
         i_rid     <= i_arid;
         i_rresp   <= 2'b00; // Always OK for now 
         // Trick: address FIFO is read on last beat
         i_rlast   <= arff_read;
      
         read_state <= nxt_read_state;
         read_adrs  <= nxt_read_adrs;
         read_count <= nxt_read_count;
      end
   end

 
   
  /**********************************************\
   *  ___  ___                                  *
   *  |  \/  |                                  *
   *  | .  . | ___ _ __ ___   ___  _ __ _   _   *
   *  | |\/| |/ _ \ '_ ` _ \ / _ \| '__| | | |  *
   *  | |  | |  __/ | | | | | (_) | |  | |_| |  *
   *  \_|  |_/\___|_| |_| |_|\___/|_|   \__, |  *
   *                                     __/ |  *
   *                                    |___/   *           
  \**********************************************/

   
////////////////////
reg [31:0] memory [0:MEMDEPTH-1];

wire [L2MEM_DEPTH-1:0] mem_wadrs;
wire [L2MEM_DEPTH-1:0] mem_radrs;

   assign mem_wadrs = (write_state==W_IDLE) ?  i_awadrs : write_adrs;
   assign mem_radrs = (read_state==R_IDLE) ?  i_aradrs : read_adrs;

   always  @(posedge clk)
   begin
      if (mem_write)
      begin
         if (i_wstrb[0])
            memory[mem_wadrs][ 7: 0] <= i_wdata[ 7: 0];
         if (i_wstrb[1])
            memory[mem_wadrs][15: 8] <= i_wdata[15: 8];
         if (i_wstrb[2])
            memory[mem_wadrs][23:16] <= i_wdata[23:16];
         if (i_wstrb[3])
            memory[mem_wadrs][31:24] <= i_wdata[31:24];
      end
      if (mem_read)
         i_rdata <= memory[mem_radrs];
   end


//
// Direct memory access functions
// Address is direct memory location 
//
// This is to ease behavioural simulation as it allows
// zero time access to the memory for both reads and writes
//

// !!!!!!!
// Remove this when using synthesized version
// !!!!!!!!!

//synthesis translate_off
function [321:0] read;
input [L2MEM_DEPTH-1:0] address;
begin
   read = memory[address];
end
endfunction

task write;
input [L2MEM_DEPTH-1:0] address;
input [31:0] data;
begin
   memory[address] = data;
end
endtask
//synthesis translate_on
  

endmodule
   